import doctest
import os
import sys

import scippneutron

from typing import Any, Dict, Optional
from docutils.nodes import document
from sphinx.application import Sphinx
import sphinx_book_theme

sys.path.insert(0, os.path.abspath('.'))

from version import VersionInfo  # noqa: E402
# General information about the project.
project = u'scippneutron'
copyright = u'2022 Scipp contributors'
author = u'Scipp contributors'

version_info = VersionInfo(repo=project)
long_version = scippneutron.__version__
outdated = not version_info.is_latest(long_version)


def add_buttons(
    app: Sphinx,
    pagename: str,
    templatename: str,
    context: Dict[str, Any],
    doctree: Optional[document],
):
    base = "https://scipp.github.io"
    l1 = []
    l1.append({"type": "link", "text": "scipp", "url": f"{base}"})
    l1.append({"type": "link", "text": "scippnexus", "url": f"{base}/scippnexus"})
    l1.append({"type": "link", "text": "scippneutron", "url": f"{base}/scippneutron"})
    l1.append({"type": "link", "text": "ess", "url": f"{base}/ess"})
    header_buttons = context["header_buttons"]
    header_buttons.append({
        "type": "group",
        "buttons": l1,
        "icon": "fa fa-caret-down",
        "text": "Related projects"
    })
    releases = version_info.minor_releases(first='0.4')
    if outdated:
        current = f"{long_version} (outdated)"
        latest = "latest"
        entries = ['.'.join(long_version.split('.')[:2])]
    else:
        current = f"{long_version} (latest)"
        latest = f"{releases[0]} (latest)"
        entries = releases[1:]
    lines = [{"type": "link", "text": latest, "url": f"{base}/{project}"}]
    for r in entries:
        lines.append({
            "type": "link",
            "text": f"{r}",
            "url": f"{base}/{project}/release/{r}"
        })
    header_buttons.append({
        "type": "group",
        "buttons": lines,
        "icon": "fa fa-caret-down",
        "text": current
    })


sphinx_book_theme.add_launch_buttons = add_buttons

html_show_sourcelink = True

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.autosummary', 'sphinx.ext.doctest',
    'sphinx.ext.intersphinx', 'sphinx.ext.mathjax', 'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints', 'sphinx_copybutton',
    'IPython.sphinxext.ipython_directive',
    'IPython.sphinxext.ipython_console_highlighting', 'nbsphinx'
]

autodoc_type_aliases = {
    'DTypeLike': 'DTypeLike',
    'VariableLike': 'VariableLike',
    'MetaDataMap': 'MetaDataMap',
    'array_like': 'array_like',
}

rst_epilog = f"""
.. |SCIPPNEUTRON_RELEASE_MONTH| replace:: {os.popen("git show -s --format=%cd --date=format:'%B %Y'").read()}
.. |SCIPPNEUTRON_VERSION| replace:: {long_version}
"""  # noqa: E501

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'scipp': ('https://scipp.github.io/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    'xarray': ('https://xarray.pydata.org/en/stable/', None)
}

# autodocs includes everything, even irrelevant API internals. autosummary
# looks more suitable in the long run when the API grows.
# For a nice example see how xarray handles its API documentation.
autosummary_generate = True

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = True
napoleon_use_rtype = False
napoleon_preprocess_types = True
napoleon_type_aliases = {
    # objects without namespace: scipp
    "DataArray": "~scipp.DataArray",
    "Dataset": "~scipp.Dataset",
    "Variable": "~scipp.Variable",
    # objects without namespace: numpy
    "ndarray": "~numpy.ndarray",
}
typehints_defaults = 'comma'
typehints_use_rtype = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'
html_sourcelink_suffix = ''  # Avoid .ipynb.txt extensions in sources

# The master toctree document.
master_doc = 'index'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = u''
# The full version, including alpha/beta/rc tags.
release = u''

warning_is_error = True

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = 'sphinx_book_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "logo_only": True,
    "repository_url": f"https://github.com/scipp/{project}",
    "repository_branch": "main",
    "path_to_docs": "docs",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
    "show_toc_level": 2,  # Show subheadings in secondary sidebar
}

if outdated:
    html_theme_options["announcement"] = (
        f"⚠️ You are viewing the documentation for an old version of {project}. "
        f"Switch to <a href='https://scipp.github.io/{project}' "
        "style='color:white;text-decoration:underline;'"
        ">latest</a> version. ⚠️")

html_logo = "_static/logo-2022.svg"
html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'scippdoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'scippneutron.tex', u'scippneutron Documentation', u'Simon Heybrock',
     'manual'),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'scippneutron', u'scippneutron Documentation', [author], 1)]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'scippneutron', u'scippneutron Documentation', author, 'scippneutron',
     'One line description of project.', 'Miscellaneous'),
]

# -- Options for Matplotlib in notebooks ----------------------------------

nbsphinx_execute_arguments = [
    "--Session.metadata=scipp_sphinx_build=True",
]

# -- Options for doctest --------------------------------------------------

doctest_global_setup = '''
import numpy as np
import scipp as sc
'''

# Using normalize whitespace because many __str__ functions in scipp produce
# extraneous empty lines and it would look strange to include them in the docs.
doctest_default_flags = doctest.ELLIPSIS | doctest.IGNORE_EXCEPTION_DETAIL | \
                        doctest.DONT_ACCEPT_TRUE_FOR_1 | \
                        doctest.NORMALIZE_WHITESPACE

# -- Options for linkcheck ------------------------------------------------

linkcheck_ignore = [
    # Specific lines in GitHub blobs cannot be found by linkcheck.
    r'https?://github\.com/.*?/blob/[a-f0-9]+/.+?#',
    # Anchors in README's on GitHub cannot be found by linkcheck.
    r'https?://github\.com/[^/]+/[^/]+\#',
    # Many links for PRs from our release notes. Slow and unlikely to cause issues.
    'https://github.com/scipp/scipp/pull/[0-9]+',
]



import scipp as sc
import numpy as np
from .nexus_helpers import NexusBuilder, Source, Stream
import pytest
from scippneutron.file_loading.load_nexus import _load_nexus_json
"""
Many tests for load_nexus_json() are in test_load_nexus
as they are parameterised to run the same checks against
load_nexus() and load_nexus_json().
Tests in this module are for features specific to the json
representation, for example the "stream" objects which link
the data in the json template to data available to
be streamed from Kafka during an experiment.
"""


@pytest.mark.skip("TODO Stream handling with log not implemented")
def test_stream_object_as_transformation_results_in_warning():
    builder = NexusBuilder()
    builder.add_component(Source("source"))
    stream_path = "/entry/streamed_nxlog_transform"
    builder.add_stream(Stream(stream_path))
    builder.add_dataset_at_path("/entry/source/depends_on", stream_path, {})

    with pytest.warns(UserWarning):
        loaded_data, _ = _load_nexus_json(builder.json_string)

    # A 0 distance translation is used in place of the streamed transformation
    default = [0, 0, 0]
    assert np.allclose(loaded_data["source_position"].values, default)
    assert loaded_data["source_position"].unit == sc.Unit("m")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock

# flake8: noqa

import importlib.metadata
try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

from .beamline_components import position, source_position, sample_position, incident_beam, scattered_beam, Ltotal, L1, L2, two_theta
from .core import convert
from .mantid import from_mantid, to_mantid, load, fit
from .instrument_view import instrument_view
from .file_loading.load_nexus import load_nexus, load_nexus_json
from .data_streaming.data_stream import data_stream
from . import data



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st
import numpy as np
import pytest
import scipp.constants as const
from scipp.testing import strategies as scst
import scipp as sc

from scippneutron.conversion import tof as tof_conv

global_settings = {
    'max_examples': 20,
    # Generating variables needs a lot of memory. Nothing we can fix in scippneutron.
    'suppress_health_check': [HealthCheck.data_too_large]
}


def element_args():
    return {
        'allow_nan': False,
        'allow_infinity': False,
        'allow_subnormal': False,
        'min_value': 1e-2,
        'max_value': 1e6
    }


def simple_variables(dims, unit):
    return scst.variables(sizes=st.dictionaries(keys=dims,
                                                values=st.integers(min_value=1,
                                                                   max_value=4),
                                                min_size=1,
                                                max_size=1),
                          unit=unit,
                          with_variances=False,
                          dtype='float',
                          elements=element_args())


def n_simple_variables(n, dims, unit):
    return scst.n_variables(n,
                            sizes=st.dictionaries(keys=dims,
                                                  values=st.integers(min_value=1,
                                                                     max_value=4),
                                                  min_size=1,
                                                  max_size=1),
                            unit=unit,
                            with_variances=False,
                            dtype='float',
                            elements=element_args())


def time_variables():
    return simple_variables(dims=st.sampled_from(('time', 't', 'tof')),
                            unit=st.sampled_from(('s', 'ms', 'us')))


def space_variables():
    return simple_variables(dims=st.sampled_from(('position', 'spectrum', 'x')),
                            unit=st.sampled_from(('m', 'cm', 'angstrom')))


def n_space_variables(n):
    return n_simple_variables(n,
                              dims=st.sampled_from(('position', 'spectrum', 'x')),
                              unit=st.sampled_from(('m', 'cm', 'angstrom')))


def angle_variables():
    return simple_variables(dims=st.sampled_from(('angle', 'two_theta', 'alpha')),
                            unit=st.sampled_from(('deg', 'rad')))


def energy_variables():
    return simple_variables(dims=st.sampled_from(('energy', 'E', 'energy_transfer')),
                            unit=st.sampled_from(('meV', 'J')))


@given(tof=time_variables(), Ltotal=space_variables())
@settings(**global_settings)
def test_wavelength_from_tof(tof, Ltotal):
    wavelength = tof_conv.wavelength_from_tof(tof=tof, Ltotal=Ltotal)
    assert sc.allclose(wavelength,
                       sc.to_unit(const.h * tof / const.m_n / Ltotal, unit='angstrom'))


@pytest.mark.parametrize('tof_dtype', ('float64', 'int64', 'int32'))
@pytest.mark.parametrize('Ltotal_dtype', ('float64', 'float32', 'int64', 'int32'))
def test_wavelength_from_tof_double_precision(tof_dtype, Ltotal_dtype):
    tof = sc.scalar(1.2, unit='s', dtype=tof_dtype)
    Ltotal = sc.scalar(10.1, unit='m', dtype=Ltotal_dtype)
    assert tof_conv.wavelength_from_tof(tof=tof, Ltotal=Ltotal).dtype == 'float64'


@pytest.mark.parametrize('Ltotal_dtype', ('float64', 'float32', 'int64', 'int32'))
def test_wavelength_from_tof_single_precision(Ltotal_dtype):
    tof = sc.scalar(1.2, unit='s', dtype='float32')
    Ltotal = sc.scalar(10.1, unit='m', dtype=Ltotal_dtype)
    assert tof_conv.wavelength_from_tof(tof=tof, Ltotal=Ltotal).dtype == 'float32'


@given(tof=time_variables(),
       Ltotal_and_two_theta=n_space_variables(2),
       two_theta_unit=st.sampled_from(('deg', 'rad')))
@settings(**global_settings)
def test_dspacing_from_tof(tof, Ltotal_and_two_theta, two_theta_unit):
    Ltotal, two_theta = Ltotal_and_two_theta
    two_theta.unit = two_theta_unit
    dspacing = tof_conv.dspacing_from_tof(tof=tof, Ltotal=Ltotal, two_theta=two_theta)
    assert sc.allclose(
        dspacing,
        sc.to_unit(const.h * tof / const.m_n / Ltotal / 2 / sc.sin(two_theta / 2),
                   unit='angstrom'))


@pytest.mark.parametrize('tof_dtype', ('float64', 'int64', 'int32'))
@pytest.mark.parametrize('Ltotal_dtype', ('float64', 'float32', 'int64', 'int32'))
@pytest.mark.parametrize('two_theta_dtype', ('float64', 'float32', 'int64', 'int32'))
def test_dspacing_from_tof_double_precision(tof_dtype, Ltotal_dtype, two_theta_dtype):
    tof = sc.scalar(52.0, unit='s', dtype=tof_dtype)
    Ltotal = sc.scalar(0.341, unit='m', dtype=Ltotal_dtype)
    two_theta = sc.scalar(1.68, unit='rad', dtype=two_theta_dtype)
    assert tof_conv.dspacing_from_tof(tof=tof, Ltotal=Ltotal,
                                      two_theta=two_theta).dtype == 'float64'


@pytest.mark.parametrize('Ltotal_dtype', ('float64', 'float32', 'int64', 'int32'))
@pytest.mark.parametrize('two_theta_dtype', ('float64', 'float32', 'int64', 'int32'))
def test_dspacing_from_tof_single_precision(Ltotal_dtype, two_theta_dtype):
    tof = sc.scalar(52.0, unit='s', dtype='float32')
    Ltotal = sc.scalar(0.341, unit='m', dtype=Ltotal_dtype)
    two_theta = sc.scalar(1.68, unit='rad', dtype=two_theta_dtype)
    assert tof_conv.dspacing_from_tof(tof=tof, Ltotal=Ltotal,
                                      two_theta=two_theta).dtype == 'float32'


@given(tof=time_variables(), Ltotal=space_variables())
@settings(**global_settings)
def test_energy_from_tof(tof, Ltotal):
    energy = tof_conv.energy_from_tof(tof=tof, Ltotal=Ltotal)
    assert sc.allclose(energy, sc.to_unit(const.m_n * Ltotal**2 / 2 / tof**2,
                                          unit='meV'))


@pytest.mark.parametrize('tof_dtype', ('float64', 'int64'))
@pytest.mark.parametrize('Ltotal_dtype', ('float64', 'float32', 'int64'))
def test_energy_from_tof_double_precision(tof_dtype, Ltotal_dtype):
    tof = sc.scalar(478.9, unit='s', dtype=tof_dtype)
    Ltotal = sc.scalar(1.256, unit='m', dtype=Ltotal_dtype)
    assert tof_conv.energy_from_tof(tof=tof, Ltotal=Ltotal).dtype == 'float64'


@pytest.mark.parametrize('Ltotal_dtype', ('float64', 'float32', 'int64'))
def test_energy_from_tof_single_precision(Ltotal_dtype):
    tof = sc.scalar(478.9, unit='s', dtype='float32')
    Ltotal = sc.scalar(1.256, unit='m', dtype=Ltotal_dtype)
    assert tof_conv.energy_from_tof(tof=tof, Ltotal=Ltotal).dtype == 'float32'


@given(tof=time_variables(),
       L1_and_L2=n_space_variables(2),
       L2_unit=st.sampled_from(('m', 'mm', 'cm')),
       incident_energy=energy_variables())
@settings(**global_settings)
def test_energy_transfer_direct_from_tof(tof, L1_and_L2, L2_unit, incident_energy):
    L1, L2 = L1_and_L2
    L2.unit = L2_unit
    # Energies are always > 0. This matters for detection of invalid tof values.
    incident_energy = abs(incident_energy) * 1.0001

    energy_transfer = tof_conv.energy_transfer_direct_from_tof(
        tof=tof, L1=L1, L2=L2, incident_energy=incident_energy)

    t0 = sc.to_unit(sc.sqrt(const.m_n * L1**2 / 2 / incident_energy), tof.unit)
    expected = incident_energy - sc.to_unit(const.m_n * L2**2 / 2 /
                                            (tof - t0)**2, incident_energy.unit)
    expected = sc.where(tof >= t0, expected, sc.scalar(np.nan,
                                                       unit=incident_energy.unit))
    assert sc.allclose(energy_transfer, expected, equal_nan=True)


@given(tof=time_variables(),
       L1_and_L2=n_space_variables(2),
       L2_unit=st.sampled_from(('m', 'mm', 'cm')),
       final_energy=energy_variables())
@settings(**global_settings)
def test_energy_transfer_indirect_from_tof(tof, L1_and_L2, L2_unit, final_energy):
    L1, L2 = L1_and_L2
    L2.unit = L2_unit
    # Energies are always > 0. This matters for detection of invalid tof values.
    final_energy = abs(final_energy) * 1.0001

    energy_transfer = tof_conv.energy_transfer_indirect_from_tof(
        tof=tof, L1=L1, L2=L2, final_energy=final_energy)

    t0 = sc.to_unit(sc.sqrt(const.m_n * L2**2 / 2 / final_energy), tof.unit)
    expected = sc.to_unit(const.m_n * L1**2 / 2 /
                          (tof - t0)**2, final_energy.unit) - final_energy
    expected = sc.where(tof >= t0, expected, sc.scalar(np.nan, unit=final_energy.unit))
    assert sc.allclose(energy_transfer, expected, equal_nan=True)


@given(wavelength=space_variables())
@settings(**global_settings)
def test_energy_from_wavelength(wavelength):
    energy = tof_conv.energy_from_wavelength(wavelength=wavelength)
    assert sc.allclose(
        energy, sc.to_unit(const.h**2 / 2 / const.m_n / wavelength**2, unit='meV'))


@pytest.mark.parametrize('wavelength_dtype', ('float64', 'int64'))
def test_energy_from_wavelength_double_precision(wavelength_dtype):
    wavelength = sc.scalar(60.5, unit='m', dtype=wavelength_dtype)
    assert tof_conv.energy_from_wavelength(wavelength=wavelength).dtype == 'float64'


def test_energy_from_wavelength_single_precision():
    wavelength = sc.scalar(60.5, unit='m', dtype='float32')
    assert tof_conv.energy_from_wavelength(wavelength=wavelength).dtype == 'float32'


@given(energy=energy_variables())
@settings(**global_settings)
def test_wavelength_from_energy(energy):
    wavelength = tof_conv.wavelength_from_energy(energy=energy)
    assert sc.allclose(
        wavelength,
        sc.to_unit(const.h / sc.sqrt(2 * const.m_n * energy), unit='angstrom'))


@pytest.mark.parametrize('energy_dtype', ('float64', 'int64'))
def test_wavelength_from_energy_double_precision(energy_dtype):
    energy = sc.scalar(61.0, unit='meV', dtype=energy_dtype)
    assert tof_conv.wavelength_from_energy(energy=energy).dtype == 'float64'


def test_wavelength_from_energy_single_precision():
    energy = sc.scalar(61.0, unit='meV', dtype='float32')
    assert tof_conv.wavelength_from_energy(energy=energy).dtype == 'float32'


@given(wavelength=space_variables(), two_theta=angle_variables())
@settings(**global_settings)
def test_Q_from_wavelength(wavelength, two_theta):
    Q = tof_conv.Q_from_wavelength(wavelength=wavelength, two_theta=two_theta)
    assert sc.allclose(Q, 4 * np.pi * sc.sin(two_theta / 2) / wavelength)


@pytest.mark.parametrize('wavelength_dtype', ('float64', 'int64'))
@pytest.mark.parametrize('two_theta_dtype', ('float64', 'float32', 'int64'))
def test_Q_from_wavelength_double_precision(wavelength_dtype, two_theta_dtype):
    wavelength = sc.scalar(3.51, unit='s', dtype=wavelength_dtype)
    two_theta = sc.scalar(0.041, unit='deg', dtype=two_theta_dtype)
    assert tof_conv.Q_from_wavelength(wavelength=wavelength,
                                      two_theta=two_theta).dtype == 'float64'


@pytest.mark.parametrize('two_theta_dtype', ('float64', 'float32', 'int64'))
def test_Q_from_wavelength_single_precision(two_theta_dtype):
    wavelength = sc.scalar(3.51, unit='s', dtype='float32')
    two_theta = sc.scalar(0.041, unit='deg', dtype=two_theta_dtype)
    assert tof_conv.Q_from_wavelength(wavelength=wavelength,
                                      two_theta=two_theta).dtype == 'float32'


@given(Q=space_variables(), two_theta=angle_variables())
@settings(**global_settings)
def test_wavelength_from_Q(Q, two_theta):
    Q.unit = f'1/{Q.unit}'
    wavelength = tof_conv.wavelength_from_Q(Q=Q, two_theta=two_theta)
    assert sc.allclose(
        wavelength, sc.to_unit(4 * np.pi * sc.sin(two_theta / 2) / Q, unit='angstrom'))


@pytest.mark.parametrize('Q_dtype', ('float64', 'int64'))
@pytest.mark.parametrize('two_theta_dtype', ('float64', 'float32', 'int64'))
def test_wavelength_from_Q_double_precision(Q_dtype, two_theta_dtype):
    Q = sc.scalar(4.151, unit='1/nm', dtype=Q_dtype)
    two_theta = sc.scalar(5.71, unit='deg', dtype=two_theta_dtype)
    assert tof_conv.wavelength_from_Q(Q=Q, two_theta=two_theta).dtype == 'float64'


@pytest.mark.parametrize('two_theta_dtype', ('float64', 'float32', 'int64'))
def test_wavelength_from_Q_single_precision(two_theta_dtype):
    Q = sc.scalar(4.151, unit='1/nm', dtype='float32')
    two_theta = sc.scalar(5.71, unit='deg', dtype=two_theta_dtype)
    assert tof_conv.wavelength_from_Q(Q=Q, two_theta=two_theta).dtype == 'float32'


@given(wavelength=space_variables(), two_theta=angle_variables())
@settings(**global_settings)
def test_dspacing_from_wavelength(wavelength, two_theta):
    dspacing = tof_conv.dspacing_from_wavelength(wavelength=wavelength,
                                                 two_theta=two_theta)
    assert sc.allclose(
        dspacing, sc.to_unit(wavelength / 2 / sc.sin(two_theta / 2), unit='angstrom'))


@pytest.mark.parametrize('wavelength_dtype', ('float64', 'int64'))
@pytest.mark.parametrize('two_theta_dtype', ('float64', 'float32', 'int64'))
def test_dspacing_from_wavelength_double_precision(wavelength_dtype, two_theta_dtype):
    wavelength = sc.scalar(41.4, unit='m', dtype=wavelength_dtype)
    two_theta = sc.scalar(8.4, unit='rad', dtype=two_theta_dtype)
    assert tof_conv.dspacing_from_wavelength(wavelength=wavelength,
                                             two_theta=two_theta).dtype == 'float64'


@pytest.mark.parametrize('two_theta_dtype', ('float64', 'float32', 'int64'))
def test_dspacing_from_wavelength_single_precision(two_theta_dtype):
    wavelength = sc.scalar(41.4, unit='m', dtype='float32')
    two_theta = sc.scalar(8.4, unit='rad', dtype=two_theta_dtype)
    assert tof_conv.dspacing_from_wavelength(wavelength=wavelength,
                                             two_theta=two_theta).dtype == 'float32'


@given(energy=energy_variables(), two_theta=angle_variables())
@settings(**global_settings)
def test_dspacing_from_energy(energy, two_theta):
    dspacing = tof_conv.dspacing_from_energy(energy=energy, two_theta=two_theta)
    assert sc.allclose(
        dspacing,
        sc.to_unit(const.h / sc.sqrt(8 * const.m_n * energy) / sc.sin(two_theta / 2),
                   unit='angstrom'))


@pytest.mark.parametrize('energy_dtype', ('float64', 'int64'))
@pytest.mark.parametrize('two_theta_dtype', ('float64', 'float32', 'int64'))
def test_dspacing_from_energy_double_precision(energy_dtype, two_theta_dtype):
    energy = sc.scalar(26.90, unit='J', dtype=energy_dtype)
    two_theta = sc.scalar(1.985, unit='rad', dtype=two_theta_dtype)
    assert tof_conv.dspacing_from_energy(energy=energy,
                                         two_theta=two_theta).dtype == 'float64'


@pytest.mark.parametrize('two_theta_dtype', ('float64', 'float32', 'int64'))
def test_dspacing_from_energy_single_precision(two_theta_dtype):
    energy = sc.scalar(26.90, unit='J', dtype='float32')
    two_theta = sc.scalar(1.985, unit='rad', dtype=two_theta_dtype)
    assert tof_conv.dspacing_from_energy(energy=energy,
                                         two_theta=two_theta).dtype == 'float32'



import scipp as sc
import numpy as np


def make_dataset_with_beamline():
    positions = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [-1, 0, 0]]
    d = sc.Dataset(
        data={'a': sc.Variable(dims=['position', 'tof'], values=np.random.rand(4, 9))},
        coords={
            'tof':
            sc.Variable(dims=['tof'],
                        values=np.arange(1000.0, 1010.0),
                        unit=sc.units.us),
            'position':
            sc.vectors(dims=['position'], values=positions, unit=sc.units.m)
        })

    d.coords['source_position'] = sc.vector(value=np.array([0, 0, -10]),
                                            unit=sc.units.m)
    d.coords['sample_position'] = sc.vector(value=np.array([0, 0, 0]), unit=sc.units.m)
    return d



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

from typing import Optional

import pytest

import scipp as sc
import scippneutron as scn

try:
    import mantid.simpleapi as sapi
    import mantid.kernel as kernel
except ImportError:
    pytestmark = pytest.mark.skip('Mantid framework is unavailable')
    sapi = None
    kernel = None


def mantid_coord(scipp_coord: str) -> str:
    return {
        'tof': 'TOF',
        'wavelength': 'Wavelength',
        'dspacing': 'dSpacing',
        'energy': 'Energy',
        'energy_transfer': 'DeltaE'
    }[scipp_coord]


def make_workspace(coord: str,
                   emode: str = 'Elastic',
                   efixed: Optional[sc.Variable] = None) -> 'sapi.Workspace':
    ws = sapi.CreateSampleWorkspace(XMin=1000, NumBanks=1, StoreInADS=False)
    # Crop out spectra index 0 as has two_theta=0, gives inf d-spacing
    ws = sapi.CropWorkspace(ws, StartWorkspaceIndex=1, StoreInADS=False)
    ws = sapi.ConvertUnits(InputWorkspace=ws,
                           Target=mantid_coord(coord),
                           EMode=emode,
                           EFixed=efixed.value if efixed is not None else None,
                           StoreInADS=False)  # start in origin units
    ws.mutableRun().addProperty('deltaE-mode',
                                kernel.StringPropertyWithValue('deltaE-mode', emode),
                                '', True)
    if efixed is not None:
        ws.mutableRun().addProperty('Ei',
                                    kernel.FloatPropertyWithValue('Ei', efixed.value),
                                    str(efixed.unit), False)
    return ws


def mantid_convert_units(ws: 'sapi.Workspace',
                         target: str,
                         emode: str = 'Elastic',
                         efixed: Optional[sc.Variable] = None) -> sc.DataArray:
    out_ws = sapi.ConvertUnits(InputWorkspace=ws,
                               Target=mantid_coord(target),
                               EMode=emode,
                               EFixed=efixed.value if efixed is not None else None,
                               StoreInADS=False)
    out = scn.mantid.from_mantid(out_ws)
    # broadcast to circumvent common-bins conversion in from_mantid
    spec_shape = out.coords['spectrum'].shape
    out.coords[target] = sc.ones(dims=['spectrum'],
                                 shape=spec_shape) * out.coords[target]
    return out


def test_mantid_convert_tof_to_wavelength():
    in_ws = make_workspace('tof')
    out_mantid = mantid_convert_units(in_ws, 'wavelength')

    in_da = scn.mantid.from_mantid(in_ws)
    out_scipp = scn.convert(data=in_da, origin='tof', target='wavelength', scatter=True)

    assert sc.allclose(out_scipp.coords['wavelength'],
                       out_mantid.coords['wavelength'],
                       rtol=1e-8 * sc.units.one)
    assert sc.identical(out_scipp.coords['spectrum'], out_mantid.coords['spectrum'])


def test_mantid_convert_tof_to_dspacing():
    in_ws = make_workspace('tof')
    out_mantid = mantid_convert_units(in_ws, 'dspacing')

    in_da = scn.mantid.from_mantid(in_ws)
    out_scipp = scn.convert(data=in_da, origin='tof', target='dspacing', scatter=True)

    assert sc.allclose(out_scipp.coords['dspacing'],
                       out_mantid.coords['dspacing'],
                       rtol=1e-8 * sc.units.one)
    assert sc.identical(out_scipp.coords['spectrum'], out_mantid.coords['spectrum'])


def test_mantid_convert_tof_to_energy():
    in_ws = make_workspace('tof')
    out_mantid = mantid_convert_units(in_ws, 'energy')

    in_da = scn.mantid.from_mantid(in_ws)
    out_scipp = scn.convert(data=in_da, origin='tof', target='energy', scatter=True)

    # Mantid reverses the order of the energy dim.
    mantid_energy = sc.empty_like(out_mantid.coords['energy'])
    assert mantid_energy.dims[1] == 'energy'
    mantid_energy.values = out_mantid.coords['energy'].values[..., ::-1]

    assert sc.allclose(out_scipp.coords['energy'],
                       mantid_energy,
                       rtol=1e-7 * sc.units.one)
    assert sc.identical(out_scipp.coords['spectrum'], out_mantid.coords['spectrum'])


def test_mantid_convert_tof_to_direct_energy_transfer():
    efixed = 1000 * sc.Unit('meV')
    in_ws = make_workspace('tof', emode='Direct', efixed=efixed)
    out_mantid = mantid_convert_units(in_ws,
                                      'energy_transfer',
                                      emode='Direct',
                                      efixed=efixed)

    in_da = scn.mantid.from_mantid(in_ws)
    out_scipp = scn.convert(data=in_da,
                            origin='tof',
                            target='energy_transfer',
                            scatter=True)

    # The conversion consists of multiplications and additions, thus the relative error
    # changes with the inputs. In this case, small tof yields a large error due to
    # the 1/tof**2 factor in the conversion.
    # rtol is chosen to account for linearly changing tof in the input data.
    assert sc.allclose(
        out_scipp.coords['energy_transfer'],
        out_mantid.coords['energy_transfer'],
        rtol=sc.linspace('energy_transfer', 1e-6, 1e-10,
                         out_scipp.coords['energy_transfer'].sizes['energy_transfer']))
    assert sc.identical(out_scipp.coords['spectrum'], out_mantid.coords['spectrum'])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import pytest
import numpy as np
import scipp as sc
from scippneutron.tof import frames


def make_array(*, npixel=3, nevent=1000, frame_length=None, time_offset=None):
    frame_length = 71.0e3 * sc.Unit('us') if frame_length is None else frame_length.to(
        unit='us')
    if time_offset is None:
        time_offset = sc.array(dims=['event'],
                               values=np.random.rand(nevent) * frame_length.value,
                               unit='us')
    pixel = sc.arange(dim='event', start=0, stop=nevent) % npixel
    events = sc.DataArray(sc.ones(sizes=time_offset.sizes),
                          coords={
                              'event_time_offset': time_offset,
                              'pixel': pixel
                          })
    da = events.group(sc.arange(dim='pixel', start=0, stop=npixel, dtype=pixel.dtype))
    da.coords['L1'] = sc.scalar(value=160.0, unit='m')
    da.coords['L2'] = sc.array(dims=['pixel'], values=np.arange(npixel), unit='m')
    return da


def test_make_frames_given_tof_bins_meta_data_raises_ValueError():
    da = make_array()
    da.coords['tof'] = sc.scalar(1.0, unit='ms')
    with pytest.raises(ValueError):
        frames.make_frames(da,
                           frame_length=71.0 * sc.Unit('ms'),
                           frame_offset=30.1 * sc.Unit('ms'),
                           lambda_min=2.5 * sc.Unit('Angstrom'))


def test_make_frames_given_tof_event_meta_data_raises_ValueError():
    da = make_array()
    da.bins.coords['tof'] = da.bins.coords['event_time_offset']
    with pytest.raises(ValueError):
        frames.make_frames(da,
                           frame_length=71.0 * sc.Unit('ms'),
                           frame_offset=30.1 * sc.Unit('ms'),
                           lambda_min=2.5 * sc.Unit('Angstrom'))


def test_make_frames_no_shift_and_infinite_energy_yields_tof_equal_time_offset():
    da = make_array(frame_length=71.0 * sc.Unit('ms'))
    da = frames.make_frames(da,
                            frame_length=71.0 * sc.Unit('ms'),
                            frame_offset=0.0 * sc.Unit('ms'),
                            lambda_min=0.0 * sc.Unit('Angstrom'))
    assert sc.identical(da.bins.coords['tof'], da.bins.attrs['event_time_offset'])


def test_make_frames_no_shift_and_no_events_below_lambda_min_yields_tof_equal_time_offset(  # noqa #501
):
    da = make_array(frame_length=71.0 * sc.Unit('ms'))
    da.bins.coords['event_time_offset'] += sc.to_unit(
        10.0 * sc.Unit('ms'), da.bins.coords['event_time_offset'].bins.unit)
    da = frames.make_frames(da,
                            frame_length=81.0 * sc.Unit('ms'),
                            frame_offset=0.0 * sc.Unit('ms'),
                            lambda_min=0.2 * sc.Unit('Angstrom'))
    assert sc.identical(da.bins.coords['tof'], da.bins.attrs['event_time_offset'])


def test_make_frames_time_offset_pivot_and_min_define_frames():
    # events [before, after, after, before] pivot point
    time_offset = sc.array(dims=['event'], values=[5.0, 70.0, 21.0, 6.0], unit='ms')
    da = make_array(frame_length=71.0 * sc.Unit('ms'),
                    npixel=1,
                    nevent=4,
                    time_offset=time_offset)
    pivot = sc.to_unit(10.0 * sc.Unit('ms'),
                       da.bins.coords['event_time_offset'].bins.unit)
    da.coords['time_offset_pivot'] = pivot
    da.coords['tof_min'] = 200.0 * sc.Unit('ms')
    da = frames.make_frames(da, frame_length=71.0 * sc.Unit('ms'))
    tof = da.bins.coords['tof'].values[0]
    tof_values = [
        71.0 - 10.0 + 5.0 + 200.0,
        60.0 + 200.0,
        11.0 + 200.0,
        71.0 - 10.0 + 6.0 + 200.0,
    ]
    assert sc.identical(tof, sc.array(dims=['event'], unit='ms', values=tof_values))


def tof_array(*, npixel=3, nevent=1000, frame_length=None, tof_min=None):
    frame_length = 71.0 * sc.Unit('ms') if frame_length is None else frame_length
    tof_min = 234.0 * sc.Unit('ms') if tof_min is None else tof_min
    tof = sc.array(dims=['event'], values=np.random.rand(nevent)) * frame_length.to(
        unit=tof_min.unit) + tof_min
    pixel = sc.arange(dim='event', start=0, stop=nevent) % npixel
    events = sc.DataArray(sc.ones(sizes=tof.sizes), coords={'tof': tof, 'pixel': pixel})
    da = events.group(sc.arange(dim='pixel', start=0, stop=npixel, dtype=pixel.dtype))
    da.coords['L1'] = sc.scalar(value=160.0, unit='m')
    da.coords['L2'] = sc.array(dims=['pixel'], values=np.arange(npixel), unit='m')
    return da


def tof_to_time_offset(tof, *, frame_length, frame_offset):
    unit = tof.bins.unit
    return (frame_offset.to(unit=unit) + tof) % frame_length.to(unit=unit)


@pytest.mark.parametrize(
    "tof_min", [234.0 * sc.Unit('ms'), 37000.0 * sc.Unit('us'), 337.0 * sc.Unit('ms')])
@pytest.mark.parametrize(
    "frame_offset", [0.0 * sc.Unit('ms'), 11.0 * sc.Unit('ms'), 9999.0 * sc.Unit('us')])
def test_make_frames_reproduces_true_pulses(tof_min, frame_offset):
    from scippneutron.conversion.tof import wavelength_from_tof
    frame_length = 71.0 * sc.Unit('ms')
    # Setup data with known 'tof' coord, which will serve as a reference
    da = tof_array(frame_length=frame_length, tof_min=tof_min)
    reference = da.bins.coords['tof'].copy()
    # Compute backwards to "raw" input with 'event_time_offset'. 'tof' coord is removed
    da.bins.coords['event_time_offset'] = tof_to_time_offset(da.bins.coords.pop('tof'),
                                                             frame_length=frame_length,
                                                             frame_offset=frame_offset)
    lambda_min = wavelength_from_tof(tof=tof_min,
                                     Ltotal=da.coords['L1'] + da.coords['L2'])

    da = frames.make_frames(da,
                            frame_length=frame_length,
                            frame_offset=frame_offset,
                            lambda_min=lambda_min)

    # Should reproduce reference 'tof' within rounding errors
    assert sc.allclose(da.bins.coords['tof'],
                       reference,
                       atol=sc.scalar(1e-12, unit=reference.bins.unit),
                       rtol=sc.scalar(1e-12))


def time_zero_to_detection_frame_index(*, frame_offset, tof_min, frame_length,
                                       frame_stride, time_zero):
    """
    Return 0-based source frame index of detection frame.

    The source frames containing the time marked by tof_min receive index 0.
    The frame after that index 1, and so on, until frame_stride-1.
    """
    shift = frame_length * ((frame_offset + tof_min) // frame_length)
    return ((time_zero + shift) // (frame_length)).value % frame_stride


class Test_time_zero_to_detection_frame_index:
    tof_min = [234.0 * sc.Unit('ms'), 37.0 * sc.Unit('ms'), 337.0 * sc.Unit('ms')]
    frame_offset = [0.0 * sc.Unit('ms'), 11.0 * sc.Unit('ms'), 9.999 * sc.Unit('ms')]
    frame_length = [71.0 * sc.Unit('ms')]
    time_zero = [
        t0 * sc.Unit('ms') for t0 in [0.0, 11.0, 70.0, 71.0, 71.1, 300.0, 345.6]
    ]

    @pytest.mark.parametrize("tof_min", tof_min)
    @pytest.mark.parametrize("frame_offset", frame_offset)
    @pytest.mark.parametrize("frame_length", frame_length)
    @pytest.mark.parametrize("time_zero", time_zero)
    def test_frame_stride_1_yields_index_0(self, frame_length, frame_offset, tof_min,
                                           time_zero):
        assert time_zero_to_detection_frame_index(frame_length=frame_length,
                                                  frame_stride=1,
                                                  frame_offset=frame_offset,
                                                  tof_min=tof_min,
                                                  time_zero=time_zero) == 0

    def test_frame_stride_2_with_zero_params_yields_source_frame_index(self):
        frame_length = 71.0 * sc.Unit('ms')
        frame_offset = 0.0 * sc.Unit('ms')
        tof_min = 0.0 * sc.Unit('ms')
        params = dict(frame_length=frame_length,
                      frame_stride=2,
                      frame_offset=frame_offset,
                      tof_min=tof_min)
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=0.0 * sc.Unit('ms')) == 0
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=71.0 * sc.Unit('ms')) == 1
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=142.0 * sc.Unit('ms')) == 0
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=213.0 * sc.Unit('ms')) == 1

    def test_frame_stride_2_with_tof_min_yields_index_with_offset(self):
        frame_length = 100.0 * sc.Unit('ms')
        frame_offset = 0.0 * sc.Unit('ms')
        # Given fastest neutrons from source frame 0 arriving in third frame...
        tof_min = 317 * sc.Unit('ms')
        params = dict(frame_length=frame_length,
                      frame_stride=2,
                      frame_offset=frame_offset,
                      tof_min=tof_min)
        # ... should yield index 0 for time_zero=3*frame_length.
        # TODO What about rounding issues? Can we end up with whole frames offset by 1?
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=0.0 * sc.Unit('ms')) == 1
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=100.0 * sc.Unit('ms')) == 0
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=200.0 * sc.Unit('ms')) == 1
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=300.0 * sc.Unit('ms')) == 0

    def test_frame_stride_2_with_frame_offset_and_tof_min_yields_index_with_offset(
            self):
        frame_length = 100.0 * sc.Unit('ms')
        # Given frame_offset, fastest neutrons from source frame 0 arriving in fourth
        # frame...
        frame_offset = 110.0 * sc.Unit('ms')  # offset is 1 frame
        tof_min = 317 * sc.Unit('ms')  # offset is 3 frames
        params = dict(frame_length=frame_length,
                      frame_stride=2,
                      frame_offset=frame_offset,
                      tof_min=tof_min)
        # ... should yield index 0 for time_zero=4*frame_length.
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=0.0 * sc.Unit('ms')) == 0
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=100.0 * sc.Unit('ms')) == 1
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=200.0 * sc.Unit('ms')) == 0
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=299.0 * sc.Unit('ms')) == 0
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=300.0 * sc.Unit('ms')) == 1
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=301.0 * sc.Unit('ms')) == 1
        assert time_zero_to_detection_frame_index(**params,
                                                  time_zero=400.0 * sc.Unit('ms')) == 0



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import scipp as sc
import matplotlib.pyplot as plt
from scippneutron.tof import TimeDistanceDiagram


def default_frame_diagram(tmax=None):
    fig, ax = plt.subplots(1, 1)
    diagram = TimeDistanceDiagram(ax,
                                  tmax=300 * sc.Unit('ms') if tmax is None else tmax)
    det1 = 32.0 * sc.Unit('m')
    det2 = 40.0 * sc.Unit('m')
    lambda_min = 16.0 * sc.units.angstrom
    frame_offset = diagram.to_time(6.0 * sc.Unit('ms'))
    diagram.add_source_pulse(15 * sc.Unit('ms'))
    diagram.add_sample(distance=20.0 * sc.Unit('m'))
    diagram.add_detector(distance=det1, name='detector1')
    diagram.add_detector(distance=det2, name='detector2')
    diagram.add_neutrons(Lmax=1.05 * det2,
                         lambda_min=lambda_min,
                         time_offset=frame_offset)
    diagram.add_neutron(L=det1,
                        time_offset=frame_offset,
                        wavelength=20.0 * sc.units.angstrom,
                        label=r'$T_0^{i+1}+\Delta t$')

    props = dict(arrowstyle='-|>')
    x0 = diagram.frame_length
    x1 = diagram.frame_length + frame_offset
    m = sc.Unit('m')
    ms = sc.Unit('ms')
    diagram.annotate(r'$T_0^i$',
                     xy=(x0, 0 * m),
                     xytext=(x0 - 20 * ms, 3 * m),
                     arrowprops=props)
    diagram.annotate(r'$T_0^i+\Delta T_0$',
                     xy=(x1, 0 * m),
                     xytext=(x1 - 20 * ms, 5 * m),
                     arrowprops=props)
    diagram.annotate(r'$T_0^{i+1}+t_{\mathrm{pivot}}(\mathrm{det1})$',
                     xy=(205 * ms, det1),
                     xytext=(150 * ms, 10 * m),
                     arrowprops=props)
    diagram.annotate(r'$T_0^{i+2}+t_{\mathrm{pivot}}(\mathrm{det2})$',
                     xy=(235 * ms, det2),
                     xytext=(220 * ms, 15 * m),
                     arrowprops=props)

    # Pivot line
    diagram.add_neutron(time_offset=x1,
                        L=1.1 * det2,
                        wavelength=lambda_min,
                        color='black',
                        ls='dashed')

    L = diagram.frame_length.value
    ax.axvspan(2 * L, 3 * L, facecolor='grey', alpha=0.2)

    return fig


def frame_skipping_diagram(tmax=None):
    fig, ax = plt.subplots(1, 1)
    diagram = TimeDistanceDiagram(ax,
                                  tmax=500 * sc.Unit('ms') if tmax is None else tmax)
    det1 = 32.0 * sc.Unit('m')
    det2 = 40.0 * sc.Unit('m')
    lambda_min = 16.0 * sc.units.angstrom
    frame_offset = diagram.to_time(1.5 * sc.Unit('ms'))
    diagram.add_source_pulse()
    diagram.add_sample(distance=20.0 * sc.Unit('m'))
    diagram.add_detector(distance=det1, name='detector1')
    diagram.add_detector(distance=det2, name='detector2')
    diagram.add_neutrons(Lmax=1.05 * det2,
                         lambda_min=lambda_min,
                         time_offset=frame_offset,
                         stride=2)

    props = dict(arrowstyle='-|>')
    x0 = 2 * diagram.frame_length
    x1 = 2 * diagram.frame_length + frame_offset
    m = sc.Unit('m')
    ms = sc.Unit('ms')
    diagram.annotate(r'$T_0^i$',
                     xy=(x0, 0 * m),
                     xytext=(x0 - 20 * ms, 3 * m),
                     arrowprops=props)
    diagram.annotate(r'$T_0^i+\Delta T_0$',
                     xy=(x1, 0 * m),
                     xytext=(x1, 5 * m),
                     arrowprops=props)

    return fig



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock

# flake8: noqa
r"""
Specifics for time-of-flight neutron-scattering data reduction, including coordinate transformations.
"""

from .diagram import TimeDistanceDiagram



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen
"""Functions for computing coordinates in time-of-flight neutron scattering.

Some functions in this module use the neutron mass :math:`m_n`
and Planck constant :math:`h`.
Their values are taken from :mod:`scipp.constants`.
"""

import numpy as np
import scipp.constants as const
from scipp.typing import VariableLike
import scipp as sc

from .._utils import as_float_type, elem_dtype, elem_unit


def _common_dtype(a, b):
    """Very limited type promotion.
    Only useful to check if the combination of a and b results in
    single or double precision float.
    """
    if elem_dtype(a) == sc.DType.float32 and elem_dtype(b) == sc.DType.float32:
        return sc.DType.float32
    return sc.DType.float64


def wavelength_from_tof(*, tof: VariableLike, Ltotal: VariableLike) -> VariableLike:
    """Compute the wavelength from time-of-flight.

    The result is the de Broglie wavelength

    .. math::

        \\lambda = \\frac{h t}{m_n L_\\mathsf{total}}

    Where :math:`m_n` is the neutron mass and :math:`h` the Planck constant.

    Parameters
    ----------
    tof:
        Time-of-flight :math:`t`.
    Ltotal:
        Total beam length.

    Returns
    -------
    :
        Wavelength :math:`\\lambda`.
        Has unit ångström.
    """
    c = sc.to_unit(const.h / const.m_n,
                   sc.units.angstrom * elem_unit(Ltotal) / elem_unit(tof),
                   copy=False)
    return as_float_type(c / Ltotal, tof) * tof


def dspacing_from_tof(*, tof: VariableLike, Ltotal: VariableLike,
                      two_theta: VariableLike) -> VariableLike:
    """Compute the d-spacing from time-of-flight.

    The result is the inter-planar lattice spacing

    .. math::

        d = \\frac{h t}{m_n L_\\mathsf{total}\\; 2 \\sin \\theta}

    Where :math:`m_n` is the neutron mass and :math:`h` the Planck constant.

    Parameters
    ----------
    tof:
        Time-of-flight :math:`t`.
    Ltotal:
        Total beam length.
    two_theta:
        Scattering angle :math:`2 \\theta`.

    Returns
    -------
    :
        Inter-planar lattice spacing :math:`d`.
        Has unit ångström.

    See Also
    --------
    scippneutron.conversions.beamline:
        Definitions of ``two_theta`` and ``Ltotal``.
    """
    c = sc.to_unit(2 * const.m_n / const.h,
                   elem_unit(tof) / sc.units.angstrom / elem_unit(Ltotal),
                   copy=False)
    return 1 / as_float_type(c * Ltotal * sc.sin(two_theta / 2), tof) * tof


def _energy_constant(energy_unit: sc.Unit, tof: VariableLike, length: VariableLike):
    return sc.to_unit(const.m_n / 2,
                      energy_unit * (elem_unit(tof) / elem_unit(length))**2,
                      copy=False)


def energy_from_tof(*, tof: VariableLike, Ltotal: VariableLike) -> VariableLike:
    """Compute the neutron energy from time-of-flight.

    The result is

    .. math::

        E = \\frac{m_n L_\\mathsf{total}^2}{2 t^2}

    Where :math:`m_n` is the neutron mass.

    Parameters
    ----------
    tof:
        Time-of-flight :math:`t`.
    Ltotal:
        Total beam length.
    Returns
    -------
    :
        Neutron energy :math:`E`.
        Has unit meV.
    """
    c = _energy_constant(sc.units.meV, tof, Ltotal)
    return as_float_type(c * Ltotal**2, tof) / tof**sc.scalar(2, dtype=elem_dtype(tof))


def _energy_transfer_t0(energy, tof, length):
    dtype = _common_dtype(energy, tof)
    c = as_float_type(_energy_constant(elem_unit(energy), tof, length), energy)
    return length.astype(dtype, copy=False) * sc.sqrt(c / energy)


def energy_transfer_direct_from_tof(*, tof: VariableLike, L1: VariableLike,
                                    L2: VariableLike,
                                    incident_energy: VariableLike) -> VariableLike:
    """Compute the energy transfer in direct inelastic scattering.

    The result is

    .. math::

        \\Delta E = E_i - \\frac{m_n L_2^2}{2 {(t - t_0)}^2}

    With

    .. math::

        t_0 = \\sqrt{m_n L_1^2 / (2 E_i)}

    and :math:`m_n` the neutron mass.

    The result is ``NaN`` for unphysical points, that is, where :math:`t < t_0`.

    Parameters
    ----------
    tof:
        Time-of-flight :math:`t`.
    L1:
        Primary beam length.
    L2:
        Secondary beam length.
    incident_energy:
        Energy before scattering :math:`E_i`.

    Returns
    -------
    :
        Energy transfer :math:`\\Delta E`.
        Has the same unit as incident_energy.

    See Also
    --------
    scippneutron.conversions.tof.energy_transfer_indirect_from_tof
    """
    t0 = _energy_transfer_t0(incident_energy, tof, L1)
    c = _energy_constant(elem_unit(incident_energy), tof, L2)
    dtype = _common_dtype(incident_energy, tof)
    scale = (c * L2**2).astype(dtype, copy=False)
    delta_tof = tof - t0
    return sc.where(delta_tof <= sc.scalar(0, unit=elem_unit(delta_tof)),
                    sc.scalar(np.nan, dtype=dtype, unit=elem_unit(incident_energy)),
                    incident_energy - scale / delta_tof**2)


def energy_transfer_indirect_from_tof(*, tof: VariableLike, L1: VariableLike,
                                      L2: VariableLike,
                                      final_energy: VariableLike) -> VariableLike:
    """Compute the energy transfer in indirect inelastic scattering.

    The result is

    .. math::

        \\Delta E = \\frac{m_n L_1^2}{2 {(t - t_0)}^2} - E_f

    With

    .. math::

        t_0 = \\sqrt{m_n L_2^2 / (2 E_f)}

    and :math:`m_n` the neutron mass.

    The result is ``NaN`` for unphysical points, that is, where :math:`t < t_0`.

    Parameters
    ----------
    tof:
        Time-of-flight :math:`t`.
    L1:
        Primary beam length.
    L2:
        Secondary beam length.
    final_energy:
        Energy after scattering :math:`E_f`.

    Returns
    -------
    :
        Energy transfer :math:`\\Delta E`.
        Has the same unit as final_energy.

    See Also
    --------
    scippneutron.conversions.tof.energy_transfer_direct_from_tof
    """
    t0 = _energy_transfer_t0(final_energy, tof, L2)
    c = _energy_constant(elem_unit(final_energy), tof, L1)
    dtype = _common_dtype(final_energy, tof)
    scale = (c * L1**2).astype(dtype, copy=False)
    delta_tof = -t0 + tof  # Order chosen such that output.dims = ['spectrum', 'tof']
    return sc.where(delta_tof <= sc.scalar(0, unit=elem_unit(delta_tof)),
                    sc.scalar(np.nan, dtype=dtype, unit=elem_unit(final_energy)),
                    scale / delta_tof**2 - final_energy)


def energy_from_wavelength(*, wavelength: VariableLike) -> VariableLike:
    """Compute the neutron energy from wavelength.

    The result is

    .. math::

        E = \\frac{h^2}{2 m_n \\lambda^2}

    Where :math:`m_n` is the neutron mass and :math:`h` the Planck constant.

    Parameters
    ----------
    wavelength:
        De Broglie wavelength :math:`\\lambda`.
        Has unit meV.

    Returns
    -------
    :
        Neutron energy :math:`E`.
    """
    c = as_float_type(
        sc.to_unit(const.h**2 / 2 / const.m_n,
                   sc.units.meV * elem_unit(wavelength)**2), wavelength)
    return c / wavelength**2


def wavelength_from_energy(*, energy: VariableLike) -> VariableLike:
    """Compute the wavelength from the neutron energy.

    The result is the de Broglie wavelength

    .. math::

        \\lambda = \\frac{h}{\\sqrt{2 m_n E}}

    Where :math:`m_n` is the neutron mass and :math:`h` the Planck constant.

    Parameters
    ----------
    energy:
        Neutron energy :math:`E`.

    Returns
    -------
    :
        Wavelength :math:`\\lambda`.
        Has unit ångström.
    """
    c = as_float_type(
        sc.to_unit(const.h**2 / 2 / const.m_n,
                   sc.units.angstrom**2 * elem_unit(energy)), energy)
    return sc.sqrt(c / energy)


def _wavelength_Q_conversions(x: VariableLike, two_theta: VariableLike) -> VariableLike:
    """Convert either from Q to wavelength or vice-versa."""
    c = as_float_type(4 * const.pi, x)
    return c * sc.sin(as_float_type(two_theta, x) / 2) / x


def Q_from_wavelength(*, wavelength: VariableLike,
                      two_theta: VariableLike) -> VariableLike:
    """Compute the momentum transfer from wavelength.

    The result is

    .. math::

        Q = \\frac{4 \\pi \\sin \\theta}{\\lambda}

    Parameters
    ----------
    wavelength:
        De Broglie wavelength :math:`\\lambda`.
    two_theta:
        Scattering angle :math:`2 \\theta`.

    Returns
    -------
    :
        Momentum transfer :math:`Q`.

    See Also
    --------
    scippneutron.conversions.beamline:
        Definition of ``two_theta``.
    """
    return _wavelength_Q_conversions(wavelength, two_theta)


def wavelength_from_Q(*, Q: VariableLike, two_theta: VariableLike) -> VariableLike:
    """Compute the wavelength from momentum transfer.

    The result is the de Broglie wavelength

    .. math::

        \\lambda = \\frac{4 \\pi \\sin \\theta}{Q}

    Parameters
    ----------
    Q:
        Momentum transfer.
    two_theta:
        Scattering angle :math:`2 \\theta`.

    Returns
    -------
    :
        Wavelength :math:`\\lambda`
        Has unit ångström.

    See Also
    --------
    scippneutron.conversions.beamline:
        Definition of ``two_theta``.
    """
    return sc.to_unit(_wavelength_Q_conversions(Q, two_theta),
                      unit='angstrom',
                      copy=False)


def dspacing_from_wavelength(*, wavelength: VariableLike,
                             two_theta: VariableLike) -> VariableLike:
    """Compute the d-spacing from wavelength.

    The result is the inter-planar lattice spacing

    .. math::

        d = \\frac{\\lambda}{2 \\sin \\theta}

    Parameters
    ----------
    wavelength:
        De Broglie wavelength :math:`\\lambda`.
    two_theta:
        Scattering angle :math:`2 \\theta`.

    Returns
    -------
    :
        Inter-planar lattice spacing :math:`d`.
        Has unit ångström.

    See Also
    --------
    scippneutron.conversions.beamline:
        Definition of ``two_theta``.
    """
    c = as_float_type(
        sc.scalar(0.5).to(unit=sc.units.angstrom / elem_unit(wavelength)), wavelength)
    return c * wavelength / sc.sin(as_float_type(two_theta, wavelength) / 2)


def dspacing_from_energy(*, energy: VariableLike,
                         two_theta: VariableLike) -> VariableLike:
    """Compute the d-spacing from the neutron energy.

    The result is the inter-planar lattice spacing

    .. math::

        d = \\frac{h}{\\sqrt{8 m_n E} \\sin \\theta}

    Where :math:`m_n` is the neutron mass and :math:`h` the Planck constant.

    Parameters
    ----------
    energy:
        Neutron energy :math:`E`.
    two_theta:
        Scattering angle :math:`2 \\theta`.

    Returns
    -------
    :
        Inter-planar lattice spacing :math:`d`.
        Has unit ångström.

    See Also
    --------
    scippneutron.conversions.beamline:
        Definition of ``two_theta``.
    """
    c = as_float_type(
        sc.to_unit(const.h**2 / 8 / const.m_n,
                   sc.units.angstrom**2 * elem_unit(energy)), energy)
    return sc.sqrt(c / energy) / sc.sin(as_float_type(two_theta, energy) / 2)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

import pytest
import scippneutron as scn
import scipp as sc

try:
    import mantid.simpleapi as sapi
    import mantid.kernel as kernel
except ImportError:
    pytestmark = pytest.mark.skip('Mantid framework is unavailable')
    sapi = None
    kernel = None


@pytest.fixture
def in_ws():
    ws = sapi.CreateSampleWorkspace(SourceDistanceFromSample=10.0,
                                    BankDistanceFromSample=1.1,
                                    BankPixelWidth=2,
                                    NumBanks=1,
                                    XMax=200,
                                    StoreInADS=False)
    return ws


@pytest.fixture
def in_da(in_ws):
    return scn.mantid.from_mantid(in_ws)


def test_beamline_compute_l1(in_ws, in_da):
    out_mantid = in_ws.detectorInfo().l1() * sc.Unit('m')
    in_da = scn.mantid.from_mantid(in_ws)
    out_scipp = scn.L1(in_da)
    assert sc.allclose(out_scipp,
                       out_mantid,
                       rtol=1e-15 * sc.units.one,
                       atol=1e-15 * out_scipp.unit)


def test_beamline_compute_l2(in_ws, in_da):
    out_mantid = sc.array(
        dims=['spectrum'],
        unit='m',
        values=[in_ws.detectorInfo().l2(i) for i in range(in_ws.detectorInfo().size())])
    in_da = scn.mantid.from_mantid(in_ws)
    out_scipp = scn.L2(in_da)
    assert sc.allclose(out_scipp,
                       out_mantid,
                       rtol=1e-15 * sc.units.one,
                       atol=1e-15 * out_scipp.unit)


def test_beamline_compute_two_theta(in_ws, in_da):
    out_mantid = sc.array(dims=['spectrum'],
                          unit='rad',
                          values=[
                              in_ws.detectorInfo().twoTheta(i)
                              for i in range(in_ws.detectorInfo().size())
                          ])
    in_da = scn.mantid.from_mantid(in_ws)
    out_scipp = scn.two_theta(in_da)
    assert sc.allclose(out_scipp,
                       out_mantid,
                       rtol=1e-13 * sc.units.one,
                       atol=1e-13 * out_scipp.unit)
